;; common.scm

(module common (mk-glob-prm
				debug-port
				debug-print
				valid-str?
				if+
				lappend
				linsert
				memv-or
				string-drop-trim
				string-take-trim
				numeric-string?
				random-string
				random-passwd
				remove-keywords
				list-contains
				resolve-args
				irx-search
				irx-match
				alist->keys
				alist?
				append-to-alist
				find-ndx
				remove-keys 
				chk=? chkci=?
				k->vci k->v chk-type)
  (import scheme chicken
          matchable irregex
          srfi-1 srfi-4 srfi-13 srfi-14
          data-structures extras)
           
;;(: mk-glob-prm (* -> procedure))
;; like parameter but scope-global, not thread-local
(define (mk-glob-prm param)
  (lambda val
    (if (not (null? val))
      (set! param (car val)))
    param))

(define debug-port (mk-glob-prm (current-error-port)))

;;(: debug-print (string #!rest * -> undefined))
(define (debug-print fmt . items)
  (if (and (debug-port) (output-port? (debug-port)))
	(begin
	  (apply format (current-error-port) 
			 (string-concatenate (list "* " fmt "\n~!")) items)
	  (flush-output (current-error-port)))))

(define-syntax valid-str?
  (syntax-rules ()
    ((_ str) (and (string? str) (> (string-length str) 0)))))

;; for convenience, a "hybrid" of if and cond
;; (if+
;;   exp ...
;; (else+
;;   exp ...)
(define-syntax if+
  (syntax-rules (else+)
    ((_ bexpr ex ... (else+ ex1 ...)) (if bexpr 
                                         (begin ex ...) 
                                         (begin ex1 ...)))
    ((_ bexpr ex1 ...) (if bexpr
                         (begin
                           ex1 ...))) ))

;;(: lappend (list #!rest * -> list))
;; lappend: arg1: list  arg2... elements to append to list
;;   (lappend '(1 2 3) 4 5 "A") -> (1 2 3 4 5 "A")  
(define (lappend lst . args)
  (append lst args))

;;(: linsert (list fixnum #!rest * -> list))
(define (linsert lst where . elems)
  (if (>= (length lst) where)
    (append lst elems)
    (let ((front (take lst where))
          (back (drop lst where)))
      (append front elems back))))

;;(: memv-or (list list -> boolean))
;; #t if any elem in elem-ls is a member of lst
(define (memv-or elem-ls lst)
  (let loop ((elems elem-ls))
   (if (null? elems) 
    #f 
    (if (memv (car elems) lst) 
      #t 
      (loop (cdr elems))))))

;;(: string-drop-trim (string fixnum #!optional fixnum -> string))
(define (string-drop-trim str ndx #!optional (incr 0))
  (string-trim-both (string-drop str (+ ndx incr))))

;;(: string-take-trim (string fixnum #!optional fixnum -> string))
(define (string-take-trim str ndx #!optional (incr 0))
  (string-trim-both (string-take str (+ ndx incr))))

;;(: numeric-string? (string -> boolean))
(define (numeric-string? str)
  (define crgx-dig (irregex '(: bol (+ digit) eol)))
  (not (eqv? (irregex-search crgx-dig str) #f)))

;;===================================================================

;; RANDOM STRING procs
(define chars (map integer->char (iota 92 34)))

(define-syntax chars->char
  (syntax-rules ()
    ((_ n) (list-ref chars n))))

;;(: random-string (#!optional fixnum -> string))
(define (random-string #!optional (len 10)) 
  (string-length "!" (number->string len) "~" 
        (list->string
          (map (lambda (_) 
                 (chars->char (random 92)))
            (iota len)))))

(define passwd-char (string-append "0123456789" "+-_"
						  "abcdefghijklmnopqrstuvwxyz"
						  "ABCDEFGHIJKLMNOPQRSTUVWXYZ"))
(define pw-char-len (string-length passwd-char))

;;(: random-passwd (#!optional fixnum -> string))
(define (random-passwd #!optional (len 10))
  (let rnd-passwd-lp ((n len) (r '()))
	(if (= n 0)
		(list->string r)
		(rnd-passwd-lp (- n 1)
			(cons (string-ref passwd-char
					  (fxmod (random 
							     (inexact->exact 
								   	  (current-milliseconds)))
							 pw-char-len))
				  r)))))

;;===================================================================

;;(: remove-keywords (list -> list))
(define (remove-keywords lst)
  (let loop-kw ((ls lst) (out '()) (kvalst '()))
    (cond
      ((null? ls)
          (list out kvalst))
      ((keyword? (car ls))
        (loop-kw (cdr ls) 
                 out 
                 (cons (car ls) kvalst)))
      (else
        (loop-kw (cdr ls) (lappend out (car ls)) kvalst)))))

;; returns lists of single elems and pairs in args-- 
;;   one or both may be '()
;;(: resolve-args (list -> (list list)))
(define (resolve-args args)
  (let loop-rsv ((ls args) (rest '()) (elems '()) (pairs '()))
    (cond
      ((null? ls)
        (if (null? rest) 
          (list elems pairs)
          (loop-rsv rest '() elems pairs)))
      ((list? (car ls))
        (loop-rsv (car ls) (cdr ls) elems pairs))
      ((pair? (car ls))
        (loop-rsv (cdr ls) rest elems (lappend pairs (car ls))))
      ((keyword? (car ls))
        (loop-rsv (drop ls 2) rest elems 
                  (alist-cons (car ls) (cadr ls) pairs)))
      (else
        (loop-rsv (cdr ls) rest (lappend elems (car ls)) pairs)))))

;;(: list-contains 
;;  (forall (a) ((list a) (list a) #!optional (a a -> boolean) -> 
;;               (or boolean (list a)))))
;; returns list of elems in any-all which are present in ls
(define (list-contains any-all ls #!optional (pred equal?))
  (let loop-any ((aa any-all) (res '()))
    (cond
      ((null? aa)
        (if (null? res)
          #f
          res))
      ((member (car aa) ls pred)
        (loop-any (cdr aa) (lappend res (car aa))))
      (else
        (loop-any (cdr aa) res)))))

;;============= irregex ===============
;; irx-matcher
;; irx-search/match -> one required prm, the sre to match +
;;    optional prms (same as returned proc below and 
;;                     are used in the same way)
;;    
;;    returned value depends on num of prms proc called with.
;;    1> irx -> above matcher proc
;;    2> irx str -> values: <matcher> <num of matches>
;;    3> irx str num/symbol -> values: 
;;              <matching subexpr of irx or 
;;                  requested info item>
;;              <matcher proc> 
;;    4> irx str num/symbol pos -> values: 
;;              <matching substring or 
;;                  info item (if symbol) or
;;                  start/end pos (num)>
;;              <matcher proc>
;; 
;; <matcher proc> prms:
;;   1> str/num: string, number. or symbol/keyword
;;     string -> new string to search for matching irx
;;     number -> get submatch (substring) 
;;                   num >= 1 -> submatch in irx or #f
;;                   num = 0 -> whole match
;;     symbol/keyword -> info from prior string match
;;     
;;   2> optional - number or symbol 
;;              (data to fetch when string in 1st prm)
;;     number -> get submatch of given string (0, 1, etc)
;;     symbol/keyword -> info re: string match (per 1st prm) or
;;                       start/end position request 
;;                       
;;   3> optional - start: or end: positions (when 1st prm is string
;;              and 2nd is number, otherwise ignored)
;;
;; (: irx-matcher ((or (struct regexp) list string) #!optional string 
;;                 (or symbol fixnum symbol) fixnum -> * *))
(define (irx-matcher irx #!optional str item pos)
  (let* 
    ((cirx (if (irregex? irx)
                 irx
                 (irregex irx)))
     (m #f)
     (n 0)
     (chk-num (lambda (num)
                (and num m (> n 0) (> num 0) (<= num n))))
     (matcher 
       (lambda (str/num #!optional retrq sepos)
         (letrec
           ((get-all (lambda ()
              (let all-loop ((nn n) (r '()))
                (if (= nn 0)
                  r
                  (all-loop (- nn 1)
                            (cons (irregex-match-substring m nn)
                                  r))))))
            (get-symb-item (lambda (item)
                            (or (and (or (keyword? item) 
                                         (symbol? item))
                                     (case item
                                       ((num-match num-match: 
                                         num-matches num-matches:
                                         n n:) n)
                                       ((cirx cirx:) cirx)
                                       ((match-obj match-obj:
                                         m m:) m)
                                       ((all all:) (get-all))
                                       (else unknown-rq:)))
                                #f)))
            (get-num-match (lambda (n0 sep)
                        (if sep
                          (case sep
                            ((start start:)
                              (irregex-match-start-index m n0))
                            ((end end:)
                              (irregex-match-end-index m n0))
                            (else
                              (irregex-match-substring m n0)))
                          (irregex-match-substring m n0)))))
         (cond
           ((string? str/num)
              (set! m (irregex-search cirx str/num))
              (set! n (or (and m (irregex-match-num-submatches m))
                          0))
              (cond
                ((and (integer? retrq) (chk-num retrq))
                 (get-num-match retrq sepos))

                (else
                  (or (get-symb-item retrq)
                      (and (not retrq) n)))))
           
           ((and (integer? str/num) (chk-num str/num))
               (get-num-match str/num retrq))

           (else (get-symb-item str/num)))))))
    
    ;; body of let* ...
    (cond
      ((and str (string? str))
       (let ((r0 (matcher str)))
         ;; now m,n vars are set
         ;; item request -> immediate result + matcher
         ;; otherwise, return matcher + result (num matches)
         (if item
           (values (matcher item pos) matcher)
           (values matcher r0))))
      
      (else (values matcher #f)))))

(define irx-search irx-matcher)

(define (irx-match irx #!optional str item pos)
  (irx-matcher `(: bos ,irx eos) str item pos))

;; ====================================

;;(: alist->keys (list -> list))
(define (alist->keys a) 
  (map car a))

;;(: alist? (list -> boolean))
(define (alist? alst)
  (let loop-alst ((ls alst))
    (cond 
      ((null? ls)
        #t)
      ((pair? (car ls))
        (loop-alst (cdr ls)))
      (else 
        #f))))

;;(: append-to-alist (string * list -> list))
;; appends data to val assoc with key, or adds
;;   (key . val) to alist if key not in alist
;; returns new alist--caller saves alist
(define (append-to-alist key data alst)
  (let ((currval (alist-ref key alst string-ci=?)))
	(if (not currval)
        (alist-cons key (list data) alst)
        (alist-update! key (lappend currval data)
                       alst string-ci=?))))

;;(: find-ndx (* list #!optional (* * -> boolean) -> (or boolean fixnum)))
(define (find-ndx el lst #!optional (eq-proc eqv?))
  (let find-loop ((ls lst)  (ndx 0))
    (if (null? ls)
      #f
      (if (eq-proc el (car ls))
        ndx
        (find-loop (cdr ls) (+ ndx 1))))))

;; delete keywords (in keylist) and values from mainls 
;; Ex: (remove-keys '(b: d:) '(a 10 b: "one" c 30 d: "two" e 3)
;;       ==>  (a 10 c 30 e 3)
;;(: remove-keys (list list -> list))
(define (remove-keys keylist mainls)
  (let rm-loop ((keys keylist) (ls mainls))
    (if (null? keys)
      ls
      (let ((ndx (find-ndx (car keys) ls)))
        (if ndx
          (rm-loop (cdr keys) (append (take ls ndx) (drop ls (+ ndx 2))))
          (rm-loop (cdr keys) ls))))))

;; val k1 === c (value of an expr)
(define (chk=? k1 c)
 (let* ((tc (chk-type c))
        (tk (chk-type k1)))
   (case tc
     ((str) (string=? c 
               (if (eqv? tk 'str)
                  k1
                  (->string k1))))
     ((num) (eqv? c 
               (case tk
                   ((num sym) k1)
                   ((kword) (string->number (keyword->string k1)))
                   ((str)     (string->number k1))
                   (else #f))))
     ((sym) (eqv? c 
               (case tk
                   ((sym num)  k1)
                   ((kword) (string->symbol (keyword->string k1)))
                   ((str) (string->symbol k1))
                   (else #f))))
     ((kword) (eqv? c
                (case tk
                  ((kword) k1)
                  ((sym) (string->keyword (symbol->string  k1)))
                  ((str) (string->keyword k1))
                  ((num) (string->keyword (number->string k1)))
                  (else #f))))
     (else #f))))

(define (chkci=? k1 c)
 (let* ((tc (chk-type c))
        (tk (chk-type k1)))
   (case tc
     ((str) (string-ci=? c 
               (if (eqv? tk 'str)
                  k1
                  (->string k1))))
     ((num) (eqv? c 
               (case tk
                   ((num sym) k1)
                   ((kword) (string->number (keyword->string k1)))
                   ((str)     (string->number k1))
                   (else #f))))
     ((sym) (eqv? c 
               (case tk
                   ((sym num) k1)
                   ((kword) (string->symbol (keyword->string k1)))
                   ((str) (string->symbol k1))
                   (else #f))))
     ((kword) (eqv? c
                (case tk
                  ((kword) k1)
                  ((sym) (string->keyword (symbol->string  k1)))
                  ((str) (string->keyword k1))
                  ((num) (string->keyword (number->string k1)))
                  (else #f))))
     (else #f))))

;;(: proc-or-ls (or procedure list))
(define (proc-or-ls item)
  ( let ((l (case (chk-type item)
             ((proc) (item))
             ((pair) item)
             (else (error "k->v" "item not a proc or list: " item)))))
    (if (not (list? l))
      (error "k->v" "item is not a valid list." item)
      l)))

;;(: k->vci ((or symbol string) list -> *)) 
(define (k->vci k lst)
  (if (or (not lst) (null? lst))
    #f
    (let ((cc (assoc k (proc-or-ls lst) chkci=?)))
      (if cc
        (cdr cc)
        #f))))

;;(: k->v ((or symbol string) list -> *))
(define (k->v k lst)
  (if (or (not lst) (null? lst))
    #f
    (let ((cc (assoc k (proc-or-ls lst) chk=?)))
      (if cc
        (cdr cc)
        #f))))

;;(: chk-type (* -> symbol))
(define (chk-type obj)
  (cond
    ((procedure? obj) 'proc)
    ((vector? obj)    'vect)
    ((u8vector? obj)  'u8vect)
	((fixnum? obj)    'fixn)
	((integer? obj)   'int)
    ((number? obj)    'num)
    ((string? obj)    'str)
    ((char?   obj)    'char)
    ((keyword? obj)   'kword)
    ((symbol? obj)    'sym)
    ((port?   obj)    'port)
    ((null?   obj)    'null)
    ((boolean? obj)   'bool)
    ((pair?   obj)    'pair)
    ((list?   obj)    'list)
    ((alist?  obj)    'alist)  
    (else  'type??)))

) ;; end module common
