;; mime-parse.scm

(module mime-parse (mp-tmp-dir
					qp-decode
                    qp-encode
					get-headers
                    kvstr->kv
                    hls->conttype
                    hls->dispo
                    hls->cont-enc
                    mime-parse)
  (import scheme chicken srfi-1 srfi-13 irregex data-structures 
          files extras posix base64)
  (import common boundary mime-db)

(define mp-tmp-dir (mk-glob-prm (current-directory)))

(define quot-rgx '(: bos (* space) (? "\"") 
                     (* space) (submatch (* (~ "\"")))
                     (* space) (? "\"") (* space) eos))

(define (trim-quotes str)
  (let ((unquot (irx-search quot-rgx str 1)))
    (or unquot str)))

(define (qp-decode str)
  (string-translate* str '(("=3B" . ";")
                           ("=2C" . ",")
                           ("=3D" . "=")
                           ("=20" . " ")
                           ("=0A" . "\n")
                           ("=0D" . "\r")
                           ("=09" . "\t")
                           ("=A0" . "")
                           ("= " . "")
                           ("=\r" . "")
                           ("=\n" . ""))))

(define (qp-encode astr)
  (let* ((str0 (irregex-replace/all '(: #\=) astr "=3D"))
         (str1 (irregex-replace/all 
            '(: #\space (submatch (or #\newline eol))) str0 "=20\\1"))
         (str2 (irregex-replace/all 
            '(: #\tab (submatch (or #\newline eol))) str1 "=09\\1"))
         (str3 (irregex-replace/all '(: (submatch (~ #\newline)) #\newline)
              str2 "\\1\r\n")))
    str3))

(define-syntax ls-add-pair 
  (syntax-rules ()
    ((_ ls key val) (lappend ls (cons key val)))))

(define matcher-key (irx-search 
                      '(: bol ($ (+ (~ (or space ":")))) ":" (* space)
                              ($ (* any)) eol)))

(define matcher-continued-val (irx-search '(: bol (+ space) 
                                                  ($ (* any)) eol)))

(define (linels->headers linels)
  (let lns-hdr-lp ((lns linels) (hdrs '()) (tmpkey #f) (tmpval ""))
    (if (null? lns)
      (or (and tmpkey (ls-add-pair hdrs tmpkey tmpval))
          hdrs)
      (let* ((nextln (car lns))
             (keyln (matcher-key nextln 1)) ;; ln starts with key -> #t
             (valln (and (not keyln)
                         (matcher-continued-val nextln 1))))
        ;; nextln: only one (or neither) of 
        ;;   (keyln valln) will be #true
        (cond
         (keyln
           (let ((val (matcher-key nextln 2))
                 (newhdrs (if tmpkey
                              (ls-add-pair hdrs tmpkey tmpval)
                              hdrs)))
             ;; replacing tmpkey/val with new key/val from nextln
             ;;    nextln: 2nd irx match is val 
             (lns-hdr-lp (cdr lns) newhdrs keyln (matcher-key nextln 2))))

         (valln
           ;; keep curr key, and extend value
          (lns-hdr-lp (cdr lns) hdrs tmpkey (string-append tmpval valln)))
         
         (else
          (lns-hdr-lp (cdr lns) hdrs tmpkey tmpval)))))))

(define (get-headers inport #!optional init)
  (let-values (((lnstr rest final) (get-header-str inport init)))
	(values (linels->headers (string-split lnstr "\r\n"))
			rest
			final)))

(define irx-k=v-split '(: bos (w/nocase (* space) ($ (+ (~ "="))) "="
                                 (* space) ($ (* any))) eos))

(define (kvstr->kv kvstr)
  (let-values (((m r) (irx-search irx-k=v-split kvstr)))
    (cond 
     ((> r 0)
      (values 
       (string-trim-both (m 1))
       (string-trim-both (trim-quotes (m 2)))))

     (else
      (values "" "")))))

;; subtype = substring after main type, e.g., "text/xxxx"
(define (get-mime-subtype mimestr)
  (let ((res (string-split mimestr "/" #t)))
    (if (and res (>= (length res) 2))
        (cadr res)
        "")))

;; irx-content-type 'splits' str at ";" and matches pieces
(define irx-content-type '(: bos ($ (+ (~ ";"))) (? ";") (* space)
                             ($ (* any)) eos))

;; hls is headers list
(define (hls->conttype hls)
  (let ((res-ct (k->vci "content-type" hls)))
    (cond 
      (res-ct
        (let-values (((m r) (irx-search irx-content-type res-ct)))
        (if (> r 0)
          ;; at least one match 
          (let ((conttype (m 1)))
            (cond
             ((string-contains-ci conttype "multipart/")
              ;; multipart: get subtype (alternative|mixed)
              ;;   get boundary from val string
              (let ((subt (get-mime-subtype conttype)))
                (let-values (((k boundary) (kvstr->kv (m 2))))
                  (if (and k (string-ci=? (string-trim-both k) 
                                          "boundary"))
                      (values "multipart" subt boundary)
                      (values #f #f #f)))))

             ((string-contains-ci conttype "image/")
              (let ((subt (get-mime-subtype conttype)))
                (values "image" subt (m 2))))

             ((string-contains-ci conttype "application/")
              (let ((subt (get-mime-subtype conttype)))
                (values "application" subt (m 2))))

             ((string-contains-ci conttype "text/")
              (let ((subt (get-mime-subtype conttype)))
                (values "text" subt (m 2))))

             (else
              ;; fall-through type/subtype
              (values (m 1) (m 2) #f))))

          ;;default mime-type
          (values "text" "plain"  #f))))

     ((null? hls)
      (values #f #f #f))
     
     (else  
      (values "text" "plain"  #f)))))

(define (hls->dispo hls)
  (let ((res-dispo (k->vci "content-disposition" hls)))
    (if res-dispo
        (let-values (((m n) (irx-search irx-content-type res-dispo)))
          (cond
           ((= n 1)
            (values (m 1) #f))
           
           ((= n 2)
            ;; (m 2) may contain one or more k=v params
            (let ((m2 (string-translate* (m 2) '(("\"" . "")))))
              (let dispo-v-loop ((kvls (string-split m2 ";"))
                                 (alist '()))
                (cond 
                 ((null? kvls)
                  ;; (m 1) -> dispo type, e.g., "attachment"
                  (values (m 1) alist))

                 (else
                  (let-values (((prmk prmv)
                                (kvstr->kv (car kvls))))

                    (dispo-v-loop 
                     (cdr kvls)
                     (alist-update! (string-trim-both prmk)
                                    (trim-quotes prmv)
                                    alist string-ci=?))))))))
           (else
            (values #f #f))))
        (values #f #f))))

(define (hls->cont-enc hls)
  (let ((res-cenc (k->vci "Content-Transfer-Encoding" hls)))
    (if res-cenc
        (let-values (((m n) (irx-search irx-content-type res-cenc)))
          (if (> n 0)
              (m 1)
              #f))
        #f)))

(define-syntax remove-hyphens
  (syntax-rules ()
    ((_ str) (if (string-suffix? "--" str)
               (string-drop-right str 2)
               str))))

;; internal
;; returns ports for tmploc and toloc if necessary
;;   otherwise #f 
(define (mk-xfer-iops fname toloc tmploc)
  (let ((todir (pathname-directory toloc)))
    (if (and todir (not (directory? todir)))
        (create-directory todir 'parents)))

  (cond
   ((not (file-exists? toloc))
    (values (open-input-file tmploc)   ;; read
            (open-output-file toloc)))  ;; write
   
   (else
    (values #f #f))))

;; if header list already parsed from inport, specify the
;;   list as hdrls, else #f.  to-dir is path to send parsed *files*
;; mime-parse -> ls of ls: ( (textlines) (htmllines) (filenames) )
(define (mime-parse inport #!optional hdrls to-dir)
  (let mp-lp ((txtlns '()) (htmllns '()) (fnames '())
              (bndls '()) (m/alt #f) (hdls hdrls) (init-data ""))
   
    (let-values (((res hdr-next finres) (get-headers inport 
													 init-data)))
     (let ((hls (or hdls res)))
      
      (let-values (((ct subt bnd) (hls->conttype hls))
                   ((dispo attribs) (hls->dispo hls)))
      (let* ((e (hls->cont-enc hls))
             (enc (and  e (string-trim-both e))))
        (debug-print "ct=~a  subt=~a  bnd=~a  enc=~a" ct subt bnd enc)
        
        (cond
         ((not ct)
          (values txtlns htmllns fnames))
         
         ((string-ci=? ct "multipart")
          (let-values (((r next1 f) (clear-to-bnd inport 
												  bnd hdr-next)))
             (mp-lp txtlns htmllns fnames (cons bnd
                                            bndls) #t #f next1)))

         ((string-ci=? ct "text")
          (let ((tx txtlns)
                (htm htmllns))
            (let-values (((res next1 final) 
                          (rcv-to-bnd inport
                                      (car bndls)
                                      ct
                                      hdr-next
                                      encoding: enc
                                      outproc:
                                        (and enc
											 (string=? enc "base64")
                                             base64-decode))))
              (cond
               ((and dispo (string-ci=? dispo "form-data"))
                (let* ((name (k->vci "name" attribs))
                       (txtval (and name 
									(or 
									  (and res
										   (string-append 
											  (string-trim-both name)
											  "=" 
											  (string-trim-both res)))
									  (string-append 
									    (string-trim-both name)
										"=" )))))
                 (set! tx (lappend txtlns (remove-hyphens txtval)))))
                         
               ((string-ci=? subt "plain")
                 (set! tx (lappend txtlns (remove-hyphens res))))
               
               ((string-ci=? subt "html")
                 (set! htm (lappend htmllns (remove-hyphens res)))))
             
              (case final
                ((eof done)
                 (values tx htm fnames))

                ((more)
                 (mp-lp tx htm fnames bndls m/alt #f next1))

                ((final)
                 (if (and m/alt (not (null? (cdr bndls))))
                   (let-values (((r nxt2 f) (clear-to-bnd inport 
                                                         (cadr bndls)
                                                         next1)))
                     (mp-lp tx htm fnames (cdr bndls) #f #f nxt2))
                   (values tx htm fnames)))
                    
                (else
                 (values tx htm fnames))))))

         ((and (string-ci=? ct "application")
               (string-ci=? subt "x-www-form-urlencoded"))          
          (let* ((lenstr (k->vci "Content-Length" hls))
                 (len (and lenstr (string->number lenstr)))
                 (post-cont (and len (> len 0)
                                 (read-string len inport)))
                 (tlns (or (and post-cont (lappend txtlns post-cont))
                           txtlns)))
            (values tlns htmllns fnames)))

         (else
          ;; image, binary, form-data, etc
          ;; filename may be in attribs if there is content-dispo
          ;;  header, otherwise could be in content-type after a ';'
          ;;  that is, in <bnd> as "name=<filename>"

          (let ((fname (if attribs
                           (alist-ref "filename" attribs string-ci=?)
                           (let-values (((nm fnm) (kvstr->kv bnd)))
                             (if (and nm (string-ci=? nm "name"))
                               fnm
                               "")))))
            (if (string-null? fname)
              (let-values (((r fnm-nx1 f) (rcv-to-bnd inport 
													  (car bndls) 
													  ct hdr-next)))
                (case f
                  ((eof done) 
                   (values txtlns htmllns fnames))

                  ((more)
                   (mp-lp txtlns htmllns fnames bndls #f #f fnm-nx1))

                  ((final)
                   (mp-lp txtlns htmllns fnames (cdr bndls)
                          #f #f fnm-nx1))))
              
              ;; Essentially, if to-dir not provided, there's no
              ;; file transfer step (from tmploc -> toloc) to do.
              ;; However, if to-dir is given, we need to compute 
              ;; toloc in order to see if there is toloc/file,
              ;; and if so, don't bother to transfer again.
              ;;
              ;; The following is admittedly a little twisted.  No 
              ;; doubt there's a better/simpler way to do it--
              ;; I'll get to it at some time...
              (let* ((newfnames (lappend fnames fname))
                     (tmploc (string-append (mp-tmp-dir) "/" fname))
                     (toloc (and to-dir (string-append to-dir "/" 
													   fname))))
               (let-values (((res loc-nxt fnl)
                               (if (or (and toloc (file-exists? toloc))
                                       (file-exists? tmploc))
                          ;; after hdr-next--skipping if either file 
						  ;; already exists
                          (clear-to-bnd inport (car bndls) hdr-next)
                          ;; if files don't exists...
                          (let* ((outport (open-output-file tmploc)))
                            ;; this writes input to output file
                            ;; initlevel is under hdr-next
                            (let-values (((r0 n2 f0) 
                                          (rcv-to-bnd inport (car bndls)
                                                      ct hdr-next 
                                                      outport: outport)))
                            ;; (debug-print "r0 ~a" r0)
                            (close-output-port outport)
                            ;; finished writing tmp file -- sets locnxt
                            (values r0 n2 f0))))))
                
                (if toloc
                  (let-values (((tmpiop toiop) 
                                (mk-xfer-iops fname toloc tmploc)))
                    ;; tmpiop and toiop: both are #t or both are #f,
                    ;; hence, only need to check one of them
                    (if+ tmpiop
                      (if (and enc (string-ci=? enc "base64"))
                          ;; use nbase64 decode to decode/copy to to-dir
                          (base64-decode tmpiop toiop)
                          ;; otherwise just copy it 
                          (file-copy tmploc toloc #t))

                      (close-input-port tmpiop)
                      (close-output-port toiop)
                      (delete-file* tmploc))))

                (case fnl
                  ((done eof) 
                   (values txtlns htmllns newfnames))

                  ((more)
                   (mp-lp txtlns htmllns newfnames
                          bndls #f #f loc-nxt))

                  ((final)
                   (mp-lp txtlns htmllns newfnames
                          (cdr bndls) #f #f loc-nxt)))))))))))))))
  
)
  
